@@ -20,11 +20,12 @@ package org.apache.maven.plugin.failsafe;
  */
 
 import java.io.File;
-import java.io.FileInputStream;
-import java.io.FileOutputStream;
 import java.io.IOException;
+import java.util.Arrays;
+import java.util.Collections;
 import java.util.List;
 
+import org.apache.maven.artifact.Artifact;
 import org.apache.maven.plugin.MojoExecutionException;
 import org.apache.maven.plugin.MojoFailureException;
 import org.apache.maven.plugin.surefire.AbstractSurefireMojo;
@@ -37,8 +38,6 @@ import org.apache.maven.shared.utils.ReaderFactory;
 import org.apache.maven.shared.utils.StringUtils;
 import org.apache.maven.surefire.suite.RunResult;
 
-import static org.apache.maven.shared.utils.io.IOUtil.close;
-
 /**
  * Run integration tests using Surefire.
  *
@@ -54,6 +53,13 @@ public class IntegrationTestMojo
 
     private static final String FAILSAFE_IN_PROGRESS_CONTEXT_KEY = "failsafe-in-progress";
 
+    /**
+     * The path representing project <em>JAR</em> file, if exists; Otherwise the directory containing generated
+     * classes of the project being tested. This will be included after the test classes in the test classpath.
+     */
+    @Parameter( defaultValue = "${project.build.outputDirectory}" )
+    private File classesDirectory;
+
     /**
      * Set this to "true" to skip running integration tests, but still compile them. Its use is NOT RECOMMENDED, but
      * quite convenient on occasion.
@@ -77,8 +83,17 @@ public class IntegrationTestMojo
      * This parameter overrides the <code>includes/excludes</code> parameters, and the TestNG <code>suiteXmlFiles</code>
      * parameter.
      * <p/>
-     * since 2.7.3 You can execute a limited number of methods in the test with adding #myMethod or #my*ethod. E.g. type
+     * Since 2.7.3 You can execute a limited number of methods in the test with adding #myMethod or #my*ethod. E.g. type
      * "-Dit.test=MyTest#myMethod" <b>supported for junit 4.x and testNg</b>
+     * <br/>
+     * Since 2.19 a complex syntax is supported in one parameter (JUnit 4, JUnit 4.7+, TestNG):<br/>
+     * "-Dit.test=???IT, !Unstable*, pkg&#47;**&#47;Ci*leIT.java, *IT#test*One+testTwo?????, #fast*+slowTest"<br/>
+     * "-Dit.test=Basic*, !%regex[.*.Unstable.*], !%regex[.*.MyIT.class#one.*|two.*], %regex[#fast.*|slow.*]"<br/>
+     * <br/>
+     * The Parameterized JUnit runner <em>describes</em> test methods using an index in brackets, so the non-regex
+     * method pattern would become: <em>#testMethod[*]</em>. If using <em>@Parameters(name="{index}: fib({0})={1}")</em>
+     * and selecting the index e.g. 5 in pattern, the non-regex method pattern would become <em>#testMethod[5:*]</em>.
+     * <br/>
      */
     @Parameter( property = "it.test" )
     private String test;
@@ -163,7 +178,7 @@ public class IntegrationTestMojo
      */
     @Parameter( property = "failsafe.parallel.forcedTimeout" )
     private double parallelTestsTimeoutForcedInSeconds;
-    
+
     /**
      * A list of &lt;include> elements specifying the tests (by pattern) that should be included in testing. When not
      * specified and when the <code>test</code> parameter is not specified, the default includes will be <code><br/>
@@ -176,8 +191,16 @@ public class IntegrationTestMojo
      * <p/>
      * Each include item may also contain a comma-separated sublist of items, which will be treated as multiple
      * &nbsp;&lt;include> entries.<br/>
+     * Since 2.19 a complex syntax is supported in one parameter (JUnit 4, JUnit 4.7+, TestNG):<br/>
+     * &nbsp;&lt;include>%regex[.*[Cat|Dog].*], Basic????, !Unstable*&lt;/include><br/>
+     * &nbsp;&lt;include>%regex[.*[Cat|Dog].*], !%regex[pkg.*Slow.*.class], pkg&#47;**&#47;*Fast*.java&lt;/include><br/>
      * <p/>
-     * This parameter is ignored if the TestNG <code>suiteXmlFiles</code> parameter is specified.
+     * This parameter is ignored if the TestNG <code>suiteXmlFiles</code> parameter is specified.<br/>
+     * <br/>
+     * <em>Notice that</em> these values are relative to the directory containing generated test classes of the project
+     * being tested. This directory is declared by the parameter <code>testClassesDirectory</code> which defaults
+     * to the POM property <code>${project.build.testOutputDirectory}</code>, typically <em>src/test/java</em>
+     * unless overridden.
      */
     @Parameter
     private List<String> includes;
@@ -194,10 +217,11 @@ public class IntegrationTestMojo
 
     /**
      * By default, Surefire forks your tests using a manifest-only JAR; set this parameter to "false" to force it to
-     * launch your tests with a plain old Java classpath. (See
-     * http://maven.apache.org/plugins/maven-surefire-plugin/examples/class-loading.html for a more detailed explanation
-     * of manifest-only JARs and their benefits.)
-     * <p/>
+     * launch your tests with a plain old Java classpath. (See the
+     * <a href="http://maven.apache.org/plugins/maven-failsafe-plugin/examples/class-loading.html">
+     *     http://maven.apache.org/plugins/maven-failsafe-plugin/examples/class-loading.html</a>
+     * for a more detailed explanation of manifest-only JARs and their benefits.)
+     * <br/>
      * Beware, setting this to "false" may cause your tests to fail on Windows if your classpath is too long.
      *
      * @since 2.4.3
@@ -211,15 +235,109 @@ public class IntegrationTestMojo
     @Parameter( property = "encoding", defaultValue = "${project.reporting.outputEncoding}" )
     private String encoding;
 
-    protected void handleSummary( RunResult summary, Exception firstForkException )
-        throws MojoExecutionException, MojoFailureException
+    /**
+     * (JUnit 4+ providers)
+     * The number of times each failing test will be rerun. If set larger than 0, rerun failing tests immediately after
+     * they fail. If a failing test passes in any of those reruns, it will be marked as pass and reported as a "flake".
+     * However, all the failing attempts will be recorded.
+     */
+    @Parameter( property = "failsafe.rerunFailingTestsCount", defaultValue = "0" )
+    private int rerunFailingTestsCount;
+
+    /**
+     * (TestNG) List of &lt;suiteXmlFile> elements specifying TestNG suite xml file locations. Note that
+     * <code>suiteXmlFiles</code> is incompatible with several other parameters of this plugin, like
+     * <code>includes/excludes</code>.<br/>
+     * This parameter is ignored if the <code>test</code> parameter is specified (allowing you to run a single test
+     * instead of an entire suite).
+     *
+     * @since 2.2
+     */
+    @Parameter( property = "failsafe.suiteXmlFiles" )
+    private File[] suiteXmlFiles;
+
+    /**
+     * Defines the order the tests will be run in. Supported values are "alphabetical", "reversealphabetical", "random",
+     * "hourly" (alphabetical on even hours, reverse alphabetical on odd hours), "failedfirst", "balanced" and
+     * "filesystem".
+     * <br/>
+     * <br/>
+     * Odd/Even for hourly is determined at the time the of scanning the classpath, meaning it could change during a
+     * multi-module build.
+     * <br/>
+     * <br/>
+     * Failed first will run tests that failed on previous run first, as well as new tests for this run.
+     * <br/>
+     * <br/>
+     * Balanced is only relevant with parallel=classes, and will try to optimize the run-order of the tests reducing the
+     * overall execution time. Initially a statistics file is created and every next test run will reorder classes.
+     * <br/>
+     * <br/>
+     * Note that the statistics are stored in a file named .surefire-XXXXXXXXX beside pom.xml, and should not be checked
+     * into version control. The "XXXXX" is the SHA1 checksum of the entire surefire configuration, so different
+     * configurations will have different statistics files, meaning if you change any config settings you will re-run
+     * once before new statistics data can be established.
+     *
+     * @since 2.7
+     */
+    @Parameter( property = "failsafe.runOrder", defaultValue = "filesystem" )
+    private String runOrder;
+
+    /**
+     * A file containing include patterns. Blank lines, or lines starting with # are ignored. If {@code includes} are
+     * also specified, these patterns are appended. Example with path, simple and regex includes:<br/>
+     * &#042;&#047;test/*<br/>
+     * &#042;&#042;&#047;NotIncludedByDefault.java<br/>
+     * %regex[.*Test.*|.*Not.*]<br/>
+     */
+    @Parameter( property = "failsafe.includesFile" )
+    private File includesFile;
+
+    /**
+     * A file containing exclude patterns. Blank lines, or lines starting with # are ignored. If {@code excludes} are
+     * also specified, these patterns are appended. Example with path, simple and regex excludes:<br/>
+     * &#042;&#047;test/*<br/>
+     * &#042;&#042;&#047;DontRunTest.*<br/>
+     * %regex[.*Test.*|.*Not.*]<br/>
+     */
+    @Parameter( property = "failsafe.excludesFile" )
+    private File excludesFile;
+
+    /**
+     * Set to error/failure count in order to skip remaining tests.
+     * Due to race conditions in parallel/forked execution this may not be fully guaranteed.<br/>
+     * Enable with system property -Dfailsafe.skipAfterFailureCount=1 or any number greater than zero.
+     * Defaults to "0".<br/>
+     * See the prerequisites and limitations in documentation:<br/>
+     * <a href="http://maven.apache.org/plugins/maven-failsafe-plugin/examples/skip-after-failure.html">
+     *     http://maven.apache.org/plugins/maven-failsafe-plugin/examples/skip-after-failure.html</a>
+     *
+     * @since 2.19
+     */
+    @Parameter( property = "failsafe.skipAfterFailureCount", defaultValue = "0" )
+    private int skipAfterFailureCount;
+
+    /**
+     * After the plugin process is shutdown by sending SIGTERM signal (CTRL+C), SHUTDOWN command is received by every
+     * forked JVM. By default (shutdown=testset) forked JVM would not continue with new test which means that
+     * the current test may still continue to run.<br/>
+     * The parameter can be configured with other two values "exit" and "kill".<br/>
+     * Using "exit" forked JVM executes System.exit(1) after the plugin process has received SIGTERM signal.<br/>
+     * Using "kill" the JVM executes Runtime.halt(1) and kills itself.
+     *
+     * @since 2.19
+     */
+    @Parameter( property = "failsafe.shutdown", defaultValue = "testset" )
+    private String shutdown;
+
+    protected int getRerunFailingTestsCount()
     {
-        writeSummary( summary, firstForkException );
+        return rerunFailingTestsCount;
     }
 
     @SuppressWarnings( "unchecked" )
-    private void writeSummary( RunResult summary, Exception firstForkException )
-        throws MojoExecutionException
+    protected void handleSummary( RunResult summary, Exception firstForkException )
+        throws MojoExecutionException, MojoFailureException
     {
         File summaryFile = getSummaryFile();
         if ( !summaryFile.getParentFile().isDirectory() )
@@ -228,8 +346,6 @@ public class IntegrationTestMojo
             summaryFile.getParentFile().mkdirs();
         }
 
-        FileOutputStream fout = null;
-        FileInputStream fin = null;
         try
         {
             Object token = getPluginContext().get( FAILSAFE_IN_PROGRESS_CONTEXT_KEY );
@@ -239,11 +355,6 @@ public class IntegrationTestMojo
         {
             throw new MojoExecutionException( e.getMessage(), e );
         }
-        finally
-        {
-            close( fin );
-            close( fout );
-        }
 
         getPluginContext().put( FAILSAFE_IN_PROGRESS_CONTEXT_KEY, FAILSAFE_IN_PROGRESS_CONTEXT_KEY );
     }
@@ -252,8 +363,9 @@ public class IntegrationTestMojo
     {
         if ( StringUtils.isEmpty( encoding ) )
         {
-            getLog().warn( "File encoding has not been set, using platform encoding " + ReaderFactory.FILE_ENCODING +
-                               ", i.e. build is platform dependent!" );
+            getLog().warn( "File encoding has not been set, using platform encoding " + ReaderFactory.FILE_ENCODING
+                           + ", i.e. build is platform dependent! The file encoding for reports output files "
+                               + "should be provided by the POM property ${project.reporting.outputEncoding}." );
             return ReaderFactory.FILE_ENCODING;
         }
         else
@@ -342,9 +454,20 @@ public class IntegrationTestMojo
         this.testClassesDirectory = testClassesDirectory;
     }
 
+    /**
+     * @return Output directory, or artifact file if artifact type is "jar". If not forking the JVM, parameter
+     * {@link #useSystemClassLoader} is ignored and the {@link org.apache.maven.surefire.booter.IsolatedClassLoader} is
+     * used instead. See the resolution of {@link #getClassLoaderConfiguration() ClassLoaderConfiguration}.
+     */
     public File getClassesDirectory()
     {
-        return classesDirectory;
+        Artifact artifact = getProject().getArtifact();
+        File artifactFile = artifact.getFile();
+
+        boolean useArtifactFile = artifactFile != null && artifactFile.isFile()
+            && artifactFile.getName().toLowerCase().endsWith( ".jar" );
+
+        return useArtifactFile ? artifactFile : classesDirectory;
     }
 
     public void setClassesDirectory( File classesDirectory )
@@ -364,15 +487,6 @@ public class IntegrationTestMojo
 
     public String getTest()
     {
-        if ( StringUtils.isBlank( test ) )
-        {
-            return null;
-        }
-        int index = test.indexOf( '#' );
-        if ( index >= 0 )
-        {
-            return test.substring( 0, index );
-        }
         return test;
     }
 
@@ -381,24 +495,6 @@ public class IntegrationTestMojo
         this.test = test;
     }
 
-    /**
-     * @since 2.7.3
-     */
-    public String getTestMethod()
-    {
-        if ( StringUtils.isBlank( test ) )
-        {
-            return null;
-        }
-        int index = this.test.indexOf( '#' );
-        if ( index >= 0 )
-        {
-            return this.test.substring( index + 1, this.test.length() );
-        }
-        return null;
-    }
-
-
     public File getSummaryFile()
     {
         return summaryFile;
@@ -459,19 +555,23 @@ public class IntegrationTestMojo
         this.forkedProcessTimeoutInSeconds = forkedProcessTimeoutInSeconds;
     }
 
-    public double getParallelTestsTimeoutInSeconds() {
+    public double getParallelTestsTimeoutInSeconds()
+    {
         return parallelTestsTimeoutInSeconds;
     }
 
-    public void setParallelTestsTimeoutInSeconds( double parallelTestsTimeoutInSeconds ) {
+    public void setParallelTestsTimeoutInSeconds( double parallelTestsTimeoutInSeconds )
+    {
         this.parallelTestsTimeoutInSeconds = parallelTestsTimeoutInSeconds;
     }
 
-    public double getParallelTestsTimeoutForcedInSeconds() {
+    public double getParallelTestsTimeoutForcedInSeconds()
+    {
         return parallelTestsTimeoutForcedInSeconds;
     }
 
-    public void setParallelTestsTimeoutForcedInSeconds( double parallelTestsTimeoutForcedInSeconds ) {
+    public void setParallelTestsTimeoutForcedInSeconds( double parallelTestsTimeoutForcedInSeconds )
+    {
         this.parallelTestsTimeoutForcedInSeconds = parallelTestsTimeoutForcedInSeconds;
     }
 
@@ -518,11 +618,21 @@ public class IntegrationTestMojo
         return failIfNoSpecifiedTests;
     }
 
-    public void setFailIfNoSpecifiedTests( Boolean failIfNoSpecifiedTests )
+    public void setFailIfNoSpecifiedTests( boolean failIfNoSpecifiedTests )
     {
         this.failIfNoSpecifiedTests = failIfNoSpecifiedTests;
     }
 
+    public int getSkipAfterFailureCount()
+    {
+        return skipAfterFailureCount;
+    }
+
+    public String getShutdown()
+    {
+        return shutdown;
+    }
+
     @Override
     public List<String> getIncludes()
     {
@@ -534,4 +644,50 @@ public class IntegrationTestMojo
     {
         this.includes = includes;
     }
+
+    public File[] getSuiteXmlFiles()
+    {
+        return suiteXmlFiles.clone();
+    }
+
+    @SuppressWarnings( "UnusedDeclaration" )
+    public void setSuiteXmlFiles( File[] suiteXmlFiles )
+    {
+        this.suiteXmlFiles = suiteXmlFiles.clone();
+    }
+
+    public String getRunOrder()
+    {
+        return runOrder;
+    }
+
+    @SuppressWarnings( "UnusedDeclaration" )
+    public void setRunOrder( String runOrder )
+    {
+        this.runOrder = runOrder;
+    }
+
+    @Override
+    public File getIncludesFile()
+    {
+        return includesFile;
+    }
+
+    @Override
+    public File getExcludesFile()
+    {
+        return excludesFile;
+    }
+
+    @Override
+    protected final List<File> suiteXmlFiles()
+    {
+        return hasSuiteXmlFiles() ? Arrays.asList( suiteXmlFiles ) : Collections.<File>emptyList();
+    }
+
+    @Override
+    protected final boolean hasSuiteXmlFiles()
+    {
+        return suiteXmlFiles != null && suiteXmlFiles.length != 0;
+    }
 }
