@@ -20,6 +20,8 @@ package org.apache.maven.plugin.surefire;
  */
 
 import java.io.File;
+import java.util.Arrays;
+import java.util.Collections;
 import java.util.List;
 import org.apache.maven.plugin.MojoExecutionException;
 import org.apache.maven.plugin.MojoFailureException;
@@ -28,7 +30,6 @@ import org.apache.maven.plugins.annotations.Mojo;
 import org.apache.maven.plugins.annotations.Parameter;
 import org.apache.maven.plugins.annotations.ResolutionScope;
 import org.apache.maven.surefire.suite.RunResult;
-import org.apache.maven.surefire.util.internal.StringUtils;
 
 /**
  * Run tests using Surefire.
@@ -43,6 +44,13 @@ public class SurefirePlugin
     implements SurefireReportParameters
 {
 
+    /**
+     * The directory containing generated classes of the project being tested. This will be included after the test
+     * classes in the test classpath.
+     */
+    @Parameter( defaultValue = "${project.build.outputDirectory}" )
+    private File classesDirectory;
+
     /**
      * Set this to "true" to ignore a failure during testing. Its use is NOT RECOMMENDED, but quite convenient on
      * occasion.
@@ -64,8 +72,17 @@ public class SurefirePlugin
      * This parameter overrides the <code>includes/excludes</code> parameters, and the TestNG <code>suiteXmlFiles</code>
      * parameter.
      * <p/>
-     * Since 2.7.3, you can execute a limited number of methods in the test by adding #myMethod or #my*ethod. For example,
-     * "-Dtest=MyTest#myMethod".  This is supported for junit 4.x and testNg.
+     * Since 2.7.3, you can execute a limited number of methods in the test by adding #myMethod or #my*ethod. For
+     * example, "-Dtest=MyTest#myMethod". This is supported for junit 4.x and testNg.<br/>
+     * <br/>
+     * Since 2.19 a complex syntax is supported in one parameter (JUnit 4, JUnit 4.7+, TestNG):<br/>
+     * "-Dtest=???Test, !Unstable*, pkg&#47;**&#47;Ci*leTest.java, *Test#test*One+testTwo?????, #fast*+slowTest"<br/>
+     * "-Dtest=Basic*, !%regex[.*.Unstable.*], !%regex[.*.MyTest.class#one.*|two.*], %regex[#fast.*|slow.*]"<br/>
+     * <br/>
+     * The Parameterized JUnit runner <em>describes</em> test methods using an index in brackets, so the non-regex
+     * method pattern would become: <em>#testMethod[*]</em>. If using <em>@Parameters(name="{index}: fib({0})={1}")</em>
+     * and selecting the index e.g. 5 in pattern, the non-regex method pattern would become <em>#testMethod[5:*]</em>.
+     * <br/>
      */
     @Parameter( property = "test" )
     private String test;
@@ -89,7 +106,6 @@ public class SurefirePlugin
     @Parameter( property = "surefire.useFile", defaultValue = "true" )
     private boolean useFile;
 
-
     /**
      * Set this to "true" to cause a failure if the none of the tests specified in -Dtest=... are run. Defaults to
      * "true".
@@ -145,7 +161,7 @@ public class SurefirePlugin
      */
     @Parameter( property = "surefire.parallel.forcedTimeout" )
     private double parallelTestsTimeoutForcedInSeconds;
-    
+
     /**
      * A list of &lt;include> elements specifying the tests (by pattern) that should be included in testing. When not
      * specified and when the <code>test</code> parameter is not specified, the default includes will be <code><br/>
@@ -158,8 +174,16 @@ public class SurefirePlugin
      * <p/>
      * Each include item may also contain a comma-separated sublist of items, which will be treated as multiple
      * &nbsp;&lt;include> entries.<br/>
+     * Since 2.19 a complex syntax is supported in one parameter (JUnit 4, JUnit 4.7+, TestNG):<br/>
+     * &nbsp;&lt;include>%regex[.*[Cat|Dog].*], Basic????, !Unstable*&lt;/include><br/>
+     * &nbsp;&lt;include>%regex[.*[Cat|Dog].*], !%regex[pkg.*Slow.*.class], pkg&#47;**&#47;*Fast*.java&lt;/include><br/>
      * <p/>
-     * This parameter is ignored if the TestNG <code>suiteXmlFiles</code> parameter is specified.
+     * This parameter is ignored if the TestNG <code>suiteXmlFiles</code> parameter is specified.<br/>
+     * <br/>
+     * <em>Notice that</em> these values are relative to the directory containing generated test classes of the project
+     * being tested. This directory is declared by the parameter <code>testClassesDirectory</code> which defaults
+     * to the POM property <code>${project.build.testOutputDirectory}</code>, typically <em>src/test/java</em>
+     * unless overridden.
      */
     @Parameter
     private List<String> includes;
@@ -176,10 +200,11 @@ public class SurefirePlugin
 
     /**
      * By default, Surefire forks your tests using a manifest-only JAR; set this parameter to "false" to force it to
-     * launch your tests with a plain old Java classpath. (See
-     * http://maven.apache.org/plugins/maven-surefire-plugin/examples/class-loading.html for a more detailed explanation
-     * of manifest-only JARs and their benefits.)
-     * <p/>
+     * launch your tests with a plain old Java classpath. (See the
+     * <a href="http://maven.apache.org/plugins/maven-surefire-plugin/examples/class-loading.html">
+     *     http://maven.apache.org/plugins/maven-surefire-plugin/examples/class-loading.html</a>
+     * for a more detailed explanation of manifest-only JARs and their benefits.)
+     * <br/>
      * Beware, setting this to "false" may cause your tests to fail on Windows if your classpath is too long.
      *
      * @since 2.4.3
@@ -187,6 +212,106 @@ public class SurefirePlugin
     @Parameter( property = "surefire.useManifestOnlyJar", defaultValue = "true" )
     private boolean useManifestOnlyJar;
 
+    /**
+     * (JUnit 4+ providers)
+     * The number of times each failing test will be rerun. If set larger than 0, rerun failing tests immediately after
+     * they fail. If a failing test passes in any of those reruns, it will be marked as pass and reported as a "flake".
+     * However, all the failing attempts will be recorded.
+     */
+    @Parameter( property = "surefire.rerunFailingTestsCount", defaultValue = "0" )
+    private int rerunFailingTestsCount;
+
+    /**
+     * (TestNG) List of &lt;suiteXmlFile> elements specifying TestNG suite xml file locations. Note that
+     * <code>suiteXmlFiles</code> is incompatible with several other parameters of this plugin, like
+     * <code>includes/excludes</code>.<br/>
+     * This parameter is ignored if the <code>test</code> parameter is specified (allowing you to run a single test
+     * instead of an entire suite).
+     *
+     * @since 2.2
+     */
+    @Parameter( property = "surefire.suiteXmlFiles" )
+    private File[] suiteXmlFiles;
+
+    /**
+     * Defines the order the tests will be run in. Supported values are "alphabetical", "reversealphabetical", "random",
+     * "hourly" (alphabetical on even hours, reverse alphabetical on odd hours), "failedfirst", "balanced" and
+     * "filesystem".
+     * <br/>
+     * <br/>
+     * Odd/Even for hourly is determined at the time the of scanning the classpath, meaning it could change during a
+     * multi-module build.
+     * <br/>
+     * <br/>
+     * Failed first will run tests that failed on previous run first, as well as new tests for this run.
+     * <br/>
+     * <br/>
+     * Balanced is only relevant with parallel=classes, and will try to optimize the run-order of the tests reducing the
+     * overall execution time. Initially a statistics file is created and every next test run will reorder classes.
+     * <br/>
+     * <br/>
+     * Note that the statistics are stored in a file named .surefire-XXXXXXXXX beside pom.xml, and should not be checked
+     * into version control. The "XXXXX" is the SHA1 checksum of the entire surefire configuration, so different
+     * configurations will have different statistics files, meaning if you change any config settings you will re-run
+     * once before new statistics data can be established.
+     *
+     * @since 2.7
+     */
+    @Parameter( property = "surefire.runOrder", defaultValue = "filesystem" )
+    private String runOrder;
+
+    /**
+     * A file containing include patterns. Blank lines, or lines starting with # are ignored. If {@code includes} are
+     * also specified, these patterns are appended. Example with path, simple and regex includes:<br/>
+     * &#042;&#047;test/*<br/>
+     * &#042;&#042;&#047;NotIncludedByDefault.java<br/>
+     * %regex[.*Test.*|.*Not.*]<br/>
+     */
+    @Parameter( property = "surefire.includesFile" )
+    private File includesFile;
+
+    /**
+     * A file containing exclude patterns. Blank lines, or lines starting with # are ignored. If {@code excludes} are
+     * also specified, these patterns are appended. Example with path, simple and regex excludes:<br/>
+     * &#042;&#047;test/*<br/>
+     * &#042;&#042;&#047;DontRunTest.*<br/>
+     * %regex[.*Test.*|.*Not.*]<br/>
+     */
+    @Parameter( property = "surefire.excludesFile" )
+    private File excludesFile;
+
+    /**
+     * Set to error/failure count in order to skip remaining tests.
+     * Due to race conditions in parallel/forked execution this may not be fully guaranteed.<br/>
+     * Enable with system property -Dsurefire.skipAfterFailureCount=1 or any number greater than zero.
+     * Defaults to "0".<br/>
+     * See the prerequisites and limitations in documentation:<br/>
+     * <a href="http://maven.apache.org/plugins/maven-surefire-plugin/examples/skip-after-failure.html">
+     *     http://maven.apache.org/plugins/maven-surefire-plugin/examples/skip-after-failure.html</a>
+     *
+     * @since 2.19
+     */
+    @Parameter( property = "surefire.skipAfterFailureCount", defaultValue = "0" )
+    private int skipAfterFailureCount;
+
+    /**
+     * After the plugin process is shutdown by sending SIGTERM signal (CTRL+C), SHUTDOWN command is received by every
+     * forked JVM. By default (shutdown=testset) forked JVM would not continue with new test which means that
+     * the current test may still continue to run.<br/>
+     * The parameter can be configured with other two values "exit" and "kill".<br/>
+     * Using "exit" forked JVM executes System.exit(1) after the plugin process has received SIGTERM signal.<br/>
+     * Using "kill" the JVM executes Runtime.halt(1) and kills itself.
+     *
+     * @since 2.19
+     */
+    @Parameter( property = "surefire.shutdown", defaultValue = "testset" )
+    private String shutdown;
+
+    protected int getRerunFailingTestsCount()
+    {
+        return rerunFailingTestsCount;
+    }
+
     protected void handleSummary( RunResult summary, Exception firstForkException )
         throws MojoExecutionException, MojoFailureException
     {
@@ -204,15 +329,6 @@ public class SurefirePlugin
         }
     }
 
-    private void assertNoFailureOrTimeout( Exception summary )
-        throws MojoFailureException
-    {
-        if ( summary != null )
-        {
-            throw new MojoFailureException( "Failure or timeout" );
-        }
-    }
-
     protected boolean isSkipExecution()
     {
         return isSkip() || isSkipTests() || isSkipExec();
@@ -318,58 +434,7 @@ public class SurefirePlugin
 
     public String getTest()
     {
-        if ( StringUtils.isBlank( test ) )
-        {
-            return null;
-        }
-        String[] testArray = StringUtils.split( test, "," );
-        StringBuilder tests = new StringBuilder();
-        for ( String aTestArray : testArray )
-        {
-            String singleTest = aTestArray;
-            int index = singleTest.indexOf( '#' );
-            if ( index >= 0 )
-            {// the way version 2.7.3.  support single test method
-                singleTest = singleTest.substring( 0, index );
-            }
-            tests.append( singleTest );
-            tests.append( "," );
-        }
-        return tests.toString();
-    }
-
-    /**
-     * @since 2.7.3
-     */
-    public String getTestMethod()
-    {
-        if ( StringUtils.isBlank( test ) )
-        {
-            return null;
-        }
-        //modified by rainLee, see http://jira.codehaus.org/browse/SUREFIRE-745
-        int index = this.test.indexOf( '#' );
-        int index2 = this.test.indexOf( "," );
-        if ( index >= 0 )
-        {
-            if ( index2 < 0 )
-            {
-                String testStrAfterFirstSharp = this.test.substring( index + 1, this.test.length() );
-                if ( !testStrAfterFirstSharp.contains( "+" ) )
-                {//the original way
-                    return testStrAfterFirstSharp;
-                }
-                else
-                {
-                    return this.test;
-                }
-            }
-            else
-            {
-                return this.test;
-            }
-        }
-        return null;
+        return test;
     }
 
     public boolean isUseSystemClassLoader()
@@ -397,11 +462,21 @@ public class SurefirePlugin
         return failIfNoSpecifiedTests;
     }
 
-    public void setFailIfNoSpecifiedTests( Boolean failIfNoSpecifiedTests )
+    public void setFailIfNoSpecifiedTests( boolean failIfNoSpecifiedTests )
     {
         this.failIfNoSpecifiedTests = failIfNoSpecifiedTests;
     }
 
+    public int getSkipAfterFailureCount()
+    {
+        return skipAfterFailureCount;
+    }
+
+    public String getShutdown()
+    {
+        return shutdown;
+    }
+
     public boolean isPrintSummary()
     {
         return printSummary;
@@ -452,19 +527,23 @@ public class SurefirePlugin
         this.forkedProcessTimeoutInSeconds = forkedProcessTimeoutInSeconds;
     }
 
-    public double getParallelTestsTimeoutInSeconds() {
+    public double getParallelTestsTimeoutInSeconds()
+    {
         return parallelTestsTimeoutInSeconds;
     }
 
-    public void setParallelTestsTimeoutInSeconds( double parallelTestsTimeoutInSeconds ) {
+    public void setParallelTestsTimeoutInSeconds( double parallelTestsTimeoutInSeconds )
+    {
         this.parallelTestsTimeoutInSeconds = parallelTestsTimeoutInSeconds;
     }
 
-    public double getParallelTestsTimeoutForcedInSeconds() {
+    public double getParallelTestsTimeoutForcedInSeconds()
+    {
         return parallelTestsTimeoutForcedInSeconds;
     }
 
-    public void setParallelTestsTimeoutForcedInSeconds( double parallelTestsTimeoutForcedInSeconds ) {
+    public void setParallelTestsTimeoutForcedInSeconds( double parallelTestsTimeoutForcedInSeconds )
+    {
         this.parallelTestsTimeoutForcedInSeconds = parallelTestsTimeoutForcedInSeconds;
     }
 
@@ -484,4 +563,50 @@ public class SurefirePlugin
     {
         this.includes = includes;
     }
+
+    public File[] getSuiteXmlFiles()
+    {
+        return suiteXmlFiles.clone();
+    }
+
+    @SuppressWarnings( "UnusedDeclaration" )
+    public void setSuiteXmlFiles( File[] suiteXmlFiles )
+    {
+        this.suiteXmlFiles = suiteXmlFiles.clone();
+    }
+
+    public String getRunOrder()
+    {
+        return runOrder;
+    }
+
+    @SuppressWarnings( "UnusedDeclaration" )
+    public void setRunOrder( String runOrder )
+    {
+        this.runOrder = runOrder;
+    }
+
+    @Override
+    public File getIncludesFile()
+    {
+        return includesFile;
+    }
+
+    @Override
+    public File getExcludesFile()
+    {
+        return excludesFile;
+    }
+
+    @Override
+    protected final List<File> suiteXmlFiles()
+    {
+        return hasSuiteXmlFiles() ? Arrays.asList( suiteXmlFiles ) : Collections.<File>emptyList();
+    }
+
+    @Override
+    protected final boolean hasSuiteXmlFiles()
+    {
+        return suiteXmlFiles != null && suiteXmlFiles.length != 0;
+    }
 }
