@@ -21,9 +21,6 @@ package org.sample.module;
 
 import static org.junit.Assert.fail;
 
-import java.lang.management.ManagementFactory;
-
-import org.junit.Before;
 import org.junit.Test;
 import org.junit.Ignore;
 
