@@ -19,8 +19,25 @@ package org.apache.maven.surefire.junitcore.pc;
  * under the License.
  */
 
+import java.lang.annotation.Annotation;
+import java.util.ArrayList;
+import java.util.Arrays;
+import java.util.Collection;
+import java.util.Collections;
+import java.util.EnumMap;
+import java.util.Iterator;
+import java.util.LinkedHashSet;
+import java.util.List;
+import java.util.Map;
+import java.util.Set;
+import java.util.concurrent.ExecutorService;
+import java.util.concurrent.Executors;
+import java.util.concurrent.ThreadFactory;
+
 import org.apache.maven.surefire.junitcore.JUnitCoreParameters;
+import org.apache.maven.surefire.report.ConsoleLogger;
 import org.apache.maven.surefire.testset.TestSetFailedException;
+import org.apache.maven.surefire.util.internal.DaemonThreadFactory;
 import org.junit.internal.runners.ErrorReportingRunner;
 import org.junit.runner.Description;
 import org.junit.runner.Runner;
@@ -32,18 +49,10 @@ import org.junit.runners.Suite;
 import org.junit.runners.model.InitializationError;
 import org.junit.runners.model.RunnerBuilder;
 
-import java.util.ArrayList;
-import java.util.Collection;
-import java.util.Collections;
-import java.util.EnumMap;
-import java.util.Iterator;
-import java.util.LinkedHashSet;
-import java.util.Map;
-import java.util.concurrent.ExecutorService;
-import java.util.concurrent.Executors;
-
-import static org.apache.maven.surefire.junitcore.pc.ParallelComputerUtil.*;
-import static org.apache.maven.surefire.junitcore.pc.Type.*;
+import static org.apache.maven.surefire.junitcore.pc.ParallelComputerUtil.resolveConcurrency;
+import static org.apache.maven.surefire.junitcore.pc.Type.CLASSES;
+import static org.apache.maven.surefire.junitcore.pc.Type.METHODS;
+import static org.apache.maven.surefire.junitcore.pc.Type.SUITES;
 
 /**
  * Executing suites, classes and methods with defined concurrency. In this example the threads which completed
@@ -60,10 +69,10 @@ import static org.apache.maven.surefire.junitcore.pc.Type.*;
  * {@link ParallelComputerBuilder#useOnePool(int)} must be greater than the number of concurrent suites and classes
  * altogether.
  * <p/>
- * The Computer can be shutdown in a separate thread. Pending tests will be interrupted if the argument is
+ * The Computer can be stopped in a separate thread. Pending tests will be interrupted if the argument is
  * <tt>true</tt>.
  * <pre>
- * computer.shutdown(true);
+ * computer.describeStopped(true);
  * </pre>
  *
  * @author Tibor Digana (tibor17)
@@ -71,10 +80,18 @@ import static org.apache.maven.surefire.junitcore.pc.Type.*;
  */
 public final class ParallelComputerBuilder
 {
+    private static final ThreadFactory DAEMON_THREAD_FACTORY = DaemonThreadFactory.newDaemonThreadFactory();
+
+    private static final Class<? extends Annotation> JCIP_NOT_THREAD_SAFE = loadNotThreadSafeAnnotations();
+
+    private static final Set<?> NULL_SINGLETON = Collections.singleton( null );
+
     static final int TOTAL_POOL_SIZE_UNDEFINED = 0;
 
     private final Map<Type, Integer> parallelGroups = new EnumMap<Type, Integer>( Type.class );
 
+    private final ConsoleLogger logger;
+
     private boolean useSeparatePools;
 
     private int totalPoolSize;
@@ -90,8 +107,9 @@ public final class ParallelComputerBuilder
      * Can be used only in unit tests.
      * Do NOT call this constructor in production.
      */
-    ParallelComputerBuilder()
+    ParallelComputerBuilder( ConsoleLogger logger )
     {
+        this.logger = logger;
         runningInTests = true;
         useSeparatePools();
         parallelGroups.put( SUITES, 0 );
@@ -99,9 +117,9 @@ public final class ParallelComputerBuilder
         parallelGroups.put( METHODS, 0 );
     }
 
-    public ParallelComputerBuilder( JUnitCoreParameters parameters )
+    public ParallelComputerBuilder( ConsoleLogger logger, JUnitCoreParameters parameters )
     {
-        this();
+        this( logger );
         runningInTests = false;
         this.parameters = parameters;
     }
@@ -214,9 +232,26 @@ public final class ParallelComputerBuilder
         return parameters == null ? 0d : parameters.getParallelTestsTimeoutForcedInSeconds();
     }
 
+    @SuppressWarnings( "unchecked" )
+    private static Class<? extends Annotation> loadNotThreadSafeAnnotations()
+    {
+        try
+        {
+            Class c = Class.forName( "net.jcip.annotations.NotThreadSafe" );
+            return c.isAnnotation() ? (Class<? extends Annotation>) c : null;
+        }
+        catch ( ClassNotFoundException e )
+        {
+            return null;
+        }
+    }
+
     final class PC
         extends ParallelComputer
     {
+        private final SingleThreadScheduler notThreadSafeTests =
+            new SingleThreadScheduler( ParallelComputerBuilder.this.logger );
+
         final Collection<ParentRunner> suites = new LinkedHashSet<ParentRunner>();
 
         final Collection<ParentRunner> nestedSuites = new LinkedHashSet<ParentRunner>();
@@ -225,7 +260,7 @@ public final class ParallelComputerBuilder
 
         final Collection<ParentRunner> nestedClasses = new LinkedHashSet<ParentRunner>();
 
-        final Collection<Runner> unscheduledRunners = new LinkedHashSet<Runner>();
+        final Collection<Runner> notParallelRunners = new LinkedHashSet<Runner>();
 
         int poolCapacity;
 
@@ -246,10 +281,29 @@ public final class ParallelComputerBuilder
         }
 
         @Override
-        public Collection<Description> shutdown( boolean shutdownNow )
+        protected ShutdownResult describeStopped( boolean shutdownNow )
         {
-            final Scheduler master = this.master;
-            return master == null ? Collections.<Description>emptyList() : master.shutdown( shutdownNow );
+            ShutdownResult shutdownResult = notThreadSafeTests.describeStopped( shutdownNow );
+            final Scheduler m = master;
+            if ( m != null )
+            {
+                ShutdownResult shutdownResultOfMaster = m.describeStopped( shutdownNow );
+                shutdownResult.getTriggeredTests().addAll( shutdownResultOfMaster.getTriggeredTests() );
+                shutdownResult.getIncompleteTests().addAll( shutdownResultOfMaster.getIncompleteTests() );
+            }
+            return shutdownResult;
+        }
+
+        @Override
+        boolean shutdownThreadPoolsAwaitingKilled()
+        {
+            boolean notInterrupted = notThreadSafeTests.shutdownThreadPoolsAwaitingKilled();
+            final Scheduler m = master;
+            if ( m != null )
+            {
+                notInterrupted &= m.shutdownThreadPoolsAwaitingKilled();
+            }
+            return notInterrupted;
         }
 
         @Override
@@ -267,7 +321,7 @@ public final class ParallelComputerBuilder
                 long suitesCount = suites.size();
                 long classesCount = classes.size() + nestedClasses.size();
                 long methodsCount = suiteClasses.embeddedChildrenCount + nestedClassesChildren;
-                if (!ParallelComputerBuilder.this.runningInTests)
+                if ( !ParallelComputerBuilder.this.runningInTests )
                 {
                     determineThreadCounts( suitesCount, classesCount, methodsCount );
                 }
@@ -276,7 +330,7 @@ public final class ParallelComputerBuilder
             }
             catch ( TestSetFailedException e )
             {
-                throw new InitializationError( e );
+                throw new InitializationError( Collections.<Throwable>singletonList( e ) );
             }
         }
 
@@ -287,7 +341,12 @@ public final class ParallelComputerBuilder
             Runner runner = super.getRunner( builder, testClass );
             if ( canSchedule( runner ) )
             {
-                if ( runner instanceof Suite )
+                if ( !isThreadSafe( runner ) )
+                {
+                    ( ( ParentRunner ) runner ).setScheduler( notThreadSafeTests.newRunnerScheduler() );
+                    notParallelRunners.add( runner );
+                }
+                else if ( runner instanceof Suite )
                 {
                     suites.add( (Suite) runner );
                 }
@@ -298,7 +357,7 @@ public final class ParallelComputerBuilder
             }
             else
             {
-                unscheduledRunners.add( runner );
+                notParallelRunners.add( runner );
             }
             return runner;
         }
@@ -314,7 +373,7 @@ public final class ParallelComputerBuilder
             allGroups.put( CLASSES, concurrency.classes );
             allGroups.put( METHODS, concurrency.methods );
             poolCapacity = concurrency.capacity;
-            splitPool &= concurrency.capacity <= 0;//fault if negative; should not happen
+            splitPool &= concurrency.capacity <= 0; // fault if negative; should not happen
         }
 
         private <T extends Runner> WrappedRunners wrapRunners( Collection<T> runners )
@@ -335,11 +394,7 @@ public final class ParallelComputerBuilder
                     }
                 }
             }
-
-            Suite wrapper = runs.isEmpty() ? null : new Suite( null, runs )
-            {
-            };
-            return new WrappedRunners( wrapper, childrenCounter );
+            return runs.isEmpty() ? new WrappedRunners() : new WrappedRunners( createSuite( runs ), childrenCounter );
         }
 
         private int countChildren( Runner runner )
@@ -352,30 +407,45 @@ public final class ParallelComputerBuilder
         private ExecutorService createPool( int poolSize )
         {
             return poolSize < Integer.MAX_VALUE
-                ? Executors.newFixedThreadPool( poolSize )
-                : Executors.newCachedThreadPool();
+                ? Executors.newFixedThreadPool( poolSize, DAEMON_THREAD_FACTORY )
+                : Executors.newCachedThreadPool( DAEMON_THREAD_FACTORY );
         }
 
         private Scheduler createMaster( ExecutorService pool, int poolSize )
         {
-            if ( !areSuitesAndClassesParallel() || poolSize <= 1 )
+            final int finalRunnersCounter = countFinalRunners(); // can be 0, 1, 2 or 3
+            final SchedulingStrategy strategy;
+            if ( finalRunnersCounter <= 1 || poolSize <= 1 )
             {
-                return new Scheduler( null, new InvokerStrategy() );
+                strategy = new InvokerStrategy( ParallelComputerBuilder.this.logger );
             }
             else if ( pool != null && poolSize == Integer.MAX_VALUE )
             {
-                return new Scheduler( null, new SharedThreadPoolStrategy( pool ) );
+                strategy = new SharedThreadPoolStrategy( ParallelComputerBuilder.this.logger, pool );
             }
             else
             {
-                return new Scheduler( null, SchedulingStrategies.createParallelStrategy( 2 ) );
+                strategy = SchedulingStrategies.createParallelStrategy( ParallelComputerBuilder.this.logger,
+                                                                        finalRunnersCounter );
             }
+            return new Scheduler( ParallelComputerBuilder.this.logger, null, strategy );
         }
 
-        private boolean areSuitesAndClassesParallel()
+        private int countFinalRunners()
         {
-            return !suites.isEmpty() && allGroups.get( SUITES ) > 0 && !classes.isEmpty()
-                && allGroups.get( CLASSES ) > 0;
+            int counter = notParallelRunners.isEmpty() ? 0 : 1;
+
+            if ( !suites.isEmpty() && allGroups.get( SUITES ) > 0 )
+            {
+                ++counter;
+            }
+
+            if ( !classes.isEmpty() && allGroups.get( CLASSES ) > 0 )
+            {
+                ++counter;
+            }
+
+            return counter;
         }
 
         private void populateChildrenFromSuites()
@@ -463,24 +533,17 @@ public final class ParallelComputerBuilder
             }
 
             // resulting runner for Computer#getSuite() scheduled by master scheduler
-            ParentRunner all = createFinalRunner( suiteSuites, suiteClasses );
+            ParentRunner all = createFinalRunner( removeNullRunners(
+                Arrays.<Runner>asList( suiteSuites, suiteClasses, createSuite( notParallelRunners ) )
+            ) );
             all.setScheduler( master );
             return all;
         }
 
-        private ParentRunner createFinalRunner( Runner... runners )
+        private ParentRunner createFinalRunner( List<Runner> runners )
             throws InitializationError
         {
-            ArrayList<Runner> all = new ArrayList<Runner>( unscheduledRunners );
-            for ( Runner runner : runners )
-            {
-                if ( runner != null )
-                {
-                    all.add( runner );
-                }
-            }
-
-            return new Suite( null, all )
+            return new Suite( null, runners )
             {
                 @Override
                 public void run( RunNotifier notifier )
@@ -515,11 +578,11 @@ public final class ParallelComputerBuilder
                 ExecutorService pool = null;
                 if ( poolSize == Integer.MAX_VALUE )
                 {
-                    pool = Executors.newCachedThreadPool();
+                    pool = Executors.newCachedThreadPool( DAEMON_THREAD_FACTORY );
                 }
                 else if ( poolSize > 0 )
                 {
-                    pool = Executors.newFixedThreadPool( poolSize );
+                    pool = Executors.newFixedThreadPool( poolSize, DAEMON_THREAD_FACTORY );
                 }
                 boolean doParallel = pool != null;
                 for ( ParentRunner runner : runners )
@@ -534,24 +597,28 @@ public final class ParallelComputerBuilder
                                            Balancer concurrency )
         {
             doParallel &= pool != null;
-            SchedulingStrategy strategy = doParallel ? new SharedThreadPoolStrategy( pool ) : new InvokerStrategy();
-            return new Scheduler( desc, master, strategy, concurrency );
+            SchedulingStrategy strategy = doParallel
+                    ? new SharedThreadPoolStrategy( ParallelComputerBuilder.this.logger, pool )
+                    : new InvokerStrategy( ParallelComputerBuilder.this.logger );
+            return new Scheduler( ParallelComputerBuilder.this.logger, desc, master, strategy, concurrency );
         }
 
         private Scheduler createScheduler( int poolSize )
         {
+            final SchedulingStrategy strategy;
             if ( poolSize == Integer.MAX_VALUE )
             {
-                return new Scheduler( null, master, SchedulingStrategies.createParallelStrategyUnbounded() );
+                strategy = SchedulingStrategies.createParallelStrategyUnbounded( ParallelComputerBuilder.this.logger );
             }
             else if ( poolSize == 0 )
             {
-                return new Scheduler( null, master, new InvokerStrategy() );
+                strategy = new InvokerStrategy( ParallelComputerBuilder.this.logger );
             }
             else
             {
-                return new Scheduler( null, master, SchedulingStrategies.createParallelStrategy( poolSize ) );
+                strategy = SchedulingStrategies.createParallelStrategy( ParallelComputerBuilder.this.logger, poolSize );
             }
+            return new Scheduler( ParallelComputerBuilder.this.logger, null, master, strategy );
         }
 
         private boolean canSchedule( Runner runner )
@@ -559,6 +626,11 @@ public final class ParallelComputerBuilder
             return !( runner instanceof ErrorReportingRunner ) && runner instanceof ParentRunner;
         }
 
+        private boolean isThreadSafe( Runner runner )
+        {
+            return runner.getDescription().getAnnotation( JCIP_NOT_THREAD_SAFE ) == null;
+        }
+
         private class SuiteFilter
             extends Filter
         {
@@ -575,15 +647,23 @@ public final class ParallelComputerBuilder
                 throws NoTestsRemainException
             {
                 super.apply( child );
-                if ( child instanceof Suite )
+                if ( child instanceof ParentRunner )
                 {
-                    nestedSuites.add( (Suite) child );
-                }
-                else if ( child instanceof ParentRunner )
-                {
-                    ParentRunner parentRunner = (ParentRunner) child;
-                    nestedClasses.add( parentRunner );
-                    nestedClassesChildren += parentRunner.getDescription().getChildren().size();
+                    ParentRunner runner = ( ParentRunner ) child;
+                    if ( !isThreadSafe( runner ) )
+                    {
+                        runner.setScheduler( notThreadSafeTests.newRunnerScheduler() );
+                    }
+                    else if ( child instanceof Suite )
+                    {
+                        nestedSuites.add( (Suite) child );
+                    }
+                    else
+                    {
+                        ParentRunner parentRunner = (ParentRunner) child;
+                        nestedClasses.add( parentRunner );
+                        nestedClassesChildren += parentRunner.getDescription().getChildren().size();
+                    }
                 }
             }
 
@@ -594,4 +674,20 @@ public final class ParallelComputerBuilder
             }
         }
     }
+
+    private static Suite createSuite( Collection<Runner> runners )
+        throws InitializationError
+    {
+        final List<Runner> onlyRunners = removeNullRunners( runners );
+        return onlyRunners.isEmpty() ? null : new Suite( null, onlyRunners )
+        {
+        };
+    }
+
+    private static List<Runner> removeNullRunners( Collection<Runner> runners )
+    {
+        final List<Runner> onlyRunners = new ArrayList<Runner>( runners );
+        onlyRunners.removeAll( NULL_SINGLETON );
+        return onlyRunners;
+    }
 }
